import asyncio
import pytest
import pytest_asyncio
import threading

from fastmcp import Client

from pydantic import BaseModel
from pymodbus.datastore import (
    ModbusSequentialDataBlock,
    ModbusServerContext,
    ModbusDeviceContext,
)
from pymodbus.server import StartAsyncTcpServer

from modbus_mcp.server import ModbusMCP


class Config(BaseModel):
    host: str = "127.0.0.1"
    port: int = 5020


async def _server_main(config: Config) -> None:
    count = 100
    store = ModbusDeviceContext(
        di=ModbusSequentialDataBlock(0, [x % 2 == 1 for x in range(count)]),
        co=ModbusSequentialDataBlock(0, [x % 2 == 0 for x in range(count)]),
        hr=ModbusSequentialDataBlock(0, list(range(0, count))),
        ir=ModbusSequentialDataBlock(0, list(range(0, count))),
    )
    context = ModbusServerContext(devices=store, single=True)
    await StartAsyncTcpServer(context, address=(config.host, config.port))


@pytest.fixture(scope="session")
def server():
    config = Config()
    thread = threading.Thread(
        target=lambda: asyncio.run(_server_main(config)), daemon=True
    )
    thread.start()
    yield config


@pytest.fixture(scope="session")
def mcp():
    return ModbusMCP()


@pytest_asyncio.fixture
async def client(mcp):
    async with Client(mcp) as c:
        yield c


@pytest.fixture()
def cli(monkeypatch):
    async def dummy_run_async(self, transport):
        return

    monkeypatch.setattr(
        "modbus_mcp.cli.ModbusMCP.run_async",
        dummy_run_async,
    )



import pytest

from fastmcp.exceptions import McpError, ToolError
from pydantic import AnyUrl
from starlette.requests import Request


@pytest.mark.asyncio
async def test_read_registers(server, mcp, client):
    """Test read_registers resource."""
    result = await client.read_resource(
        AnyUrl(f"tcp://{server.host}:{server.port}/40010?count=1&unit=1")
    )
    assert len(result) == 1
    assert result[0].text == "10"

    with pytest.raises(McpError) as e:
        await client.read_resource(AnyUrl("tcp://none:502/40010?count=1&unit=1"))
    assert "Could not read" in str(e.value)


@pytest.mark.asyncio
async def test_write_registers(server, mcp, client):
    """Test write_registers tool."""
    result = await client.call_tool(
        "write_registers",
        {
            "host": server.host,
            "port": server.port,
            "address": 40001,
            "data": [565],
            "unit": 1,
        },
    )
    assert len(result.content) == 1
    assert "succedeed" in result.content[0].text

    with pytest.raises(ToolError) as e:
        await client.call_tool(
            "write_registers",
            {
                "host": "none",
                "port": 502,
                "address": 40001,
                "data": [565],
                "unit": 1,
            },
        )
    assert "Error calling tool" in str(e.value)


@pytest.mark.asyncio
async def test_mask_write_register(server, mcp, client):
    """Test mask_write_register tool."""
    result = await client.call_tool(
        "mask_write_register",
        {
            "host": server.host,
            "port": server.port,
            "address": 40001,
            "and_mask": 0xFFFF,
            "or_mask": 0x0000,
            "unit": 1,
        },
    )
    assert len(result.content) == 1
    assert "succedeed" in result.content[0].text

    with pytest.raises(ToolError) as e:
        await client.call_tool(
            "mask_write_register",
            {
                "host": "none",
                "port": 502,
                "address": 40001,
                "and_mask": 0xFFFF,
                "or_mask": 0x0000,
                "unit": 1,
            },
        )
    assert "Error calling tool" in str(e.value)


@pytest.mark.asyncio
async def test_help_prompt(mcp, client):
    """Test help prompt."""
    result = await client.get_prompt("modbus_help", {})
    assert len(result.messages) == 5


@pytest.mark.asyncio
async def test_error_prompt(mcp, client):
    """Test error prompt."""
    result = await client.get_prompt("modbus_error", {"error": "Could not read data"})
    assert len(result.messages) == 2

    result = await client.get_prompt("modbus_error", {"error": ""})
    assert len(result.messages) == 0


@pytest.mark.asyncio
async def test_health_check(mcp):
    response = await mcp.health_check(
        Request(
            {
                "type": "http",
                "method": "GET",
                "path": "/health",
                "headers": [],
            }
        )
    )
    assert response.status_code == 200



import asyncio
import typer

from modbus_mcp.server import ModbusMCP


app = typer.Typer(
    name="modbus-mcp",
    help="ModbusMCP CLI",
)


@app.command()
def run(
    host: str | None = typer.Option(None, "--host"),
    port: int | None = typer.Option(None, "--port"),
):
    kwargs: dict[str, object] = {}
    if host is not None:
        kwargs["host"] = host
    if port is not None:
        kwargs["port"] = port
    server = ModbusMCP()
    asyncio.run(server.run_async(transport="http", **kwargs))



import pytest

from modbus_mcp.settings import Device, Settings
from modbus_mcp.utils import get_device


def test_get_device():
    settings = Settings()
    with pytest.raises(RuntimeError) as e:
        get_device(settings, "Foo")
    assert "not found" in str(e.value)
    assert get_device(settings, host="10.0.0.1", port=5020, unit=3) == (
        "10.0.0.1",
        5020,
        3,
    )
    settings.devices.append(Device(name="Foo", host="10.10.0.1", port=5020, unit=2))
    assert get_device(settings, name="Foo") == ("10.10.0.1", 5020, 2)



from fastmcp import FastMCP
from fastmcp.server.auth.providers.workos import AuthKitProvider
from fastmcp.prompts.prompt import Message
from fastmcp.resources import ResourceTemplate
from pymodbus.client import AsyncModbusTcpClient
from starlette.requests import Request
from starlette.responses import JSONResponse

from modbus_mcp.settings import Settings
from modbus_mcp.utils import get_device


settings = Settings()

_READ_FN = {
    0: ("read_coils", 1),
    1: ("read_discrete_inputs", 10001),
    3: ("read_input_registers", 30001),
    4: ("read_holding_registers", 40001),
}

_WRITE_FN = {0: ("write_coils", 1), 4: ("write_registers", 40001)}


class ModbusMCP(FastMCP):
    def __init__(self, **kwargs):
        super().__init__(
            name="Modbus MCP Server",
            auth=(
                AuthKitProvider(
                    authkit_domain=settings.auth.domain, base_url=settings.auth.url
                )
                if settings.auth.domain and settings.auth.url
                else None
            ),
            **kwargs,
        )

        self.add_template(
            ResourceTemplate.from_function(
                fn=self.read_registers,
                uri_template="tcp://{host}:{port}/{address}{?count,unit}",
            )
        )

        self.tool(
            self.read_registers,
            annotations={
                "title": "Read Registers",
                "readOnlyHint": True,
                "openWorldHint": True,
            },
        )

        self.tool(
            self.write_registers,
            annotations={
                "title": "Write Registers",
                "readOnlyHint": False,
                "openWorldHint": True,
            },
        )

        self.tool(
            self.mask_write_register,
            annotations={
                "title": "Mask Write Register",
                "readOnlyHint": False,
                "openWorldHint": True,
            },
        )

        self.prompt(self.modbus_error, name="modbus_error", tags={"modbus", "error"})
        self.prompt(self.modbus_help, name="modbus_help", tags={"modbus", "help"})

        self.custom_route("/health", methods=["GET"])(self.health_check)

    async def read_registers(
        self,
        address: int = 40001,
        count: int = 1,
        name: str | None = None,
        host: str | None = None,
        port: int | None = None,
        unit: int | None = None,
    ) -> str:
        """Reads the contents of one or more registers on a remote unit."""
        try:
            host, port, unit = get_device(settings, name, host, port, unit)
            async with AsyncModbusTcpClient(host, port=port) as client:
                func, offset = _READ_FN[address // 10000]
                method = getattr(client, func)
                res = await method(address - offset, count=count, device_id=unit)
                out = getattr(res, "registers", []) or getattr(res, "bits", [])
                return ",".join(str(x) for x in out) if count > 1 else str(out[0])
        except Exception as e:
            raise RuntimeError(
                f"Could not read {address} ({count}) from {host}:{port}"
            ) from e

    async def write_registers(
        self,
        data: list[int],
        address: int = 40001,
        name: str | None = None,
        host: str | None = None,
        port: int | None = None,
        unit: int | None = None,
    ) -> str:
        """Writes data to one or more registers on a remote unit."""
        try:
            host, port, unit = get_device(settings, name, host, port, unit)
            async with AsyncModbusTcpClient(host, port=port) as client:
                func, offset = _WRITE_FN[address // 10000]
                method = getattr(client, func)
                res = await method(address - offset, data, device_id=unit)
                if res.isError():
                    raise RuntimeError(f"Could not write to {address} on {host}:{port}")
                return f"Write to {address} on {host}:{port} has succedeed"
        except Exception as e:
            raise RuntimeError(f"{e}") from e

    async def mask_write_register(
        self,
        address: int = 40001,
        and_mask: int = 0xFFFF,
        or_mask: int = 0x0000,
        name: str | None = None,
        host: str | None = None,
        port: int | None = None,
        unit: int | None = None,
    ) -> str:
        """Mask writes data to a specified register."""
        try:
            host, port, unit = get_device(settings, name, host, port, unit)
            async with AsyncModbusTcpClient(host, port=port) as client:
                res = await client.mask_write_register(
                    address=(address - 40001),
                    and_mask=and_mask,
                    or_mask=or_mask,
                    device_id=unit,
                )
                if res.isError():
                    raise RuntimeError(
                        f"Could not mask write to {address} on {host}:{port}"
                    )
                return f"Mask write to {address} on {host}:{port} has succedeed"
        except Exception as e:
            raise RuntimeError(f"{e}") from e

    def modbus_help(self) -> list[Message]:
        """Provides examples of how to use the Modbus MCP server."""
        return [
            Message("Here are examples of how to read and write registers:"),
            Message("Please read the value of register 40001 on 127.0.0.1:502."),
            Message("Set register 40005 to 123 on host 192.168.1.10, unit 3."),
            Message("Write [1, 2, 3] to holding registers starting at address 40010."),
            Message("What is the status of input register 30010 on 10.0.0.5?"),
        ]

    def modbus_error(self, error: str | None = None) -> list[Message]:
        """Asks the user how to handle an error."""
        return (
            [
                Message(f"ERROR: {error!r}"),
                Message("Would you like to retry, change parameters, or abort?"),
            ]
            if error
            else []
        )

    async def health_check(self, request: Request) -> JSONResponse:
        return JSONResponse({"status": "ok"})
